/* $Id: input.c,v 1.17 2010/09/13 18:12:32 lamonte Exp $
 *
 * input.c
 *
 * Additional acknowledgements, news, and updates to the Qodem Project
 * are available at http://qodem.sourceforge.net .
 *
 * This module is licensed under the GNU General Public License Version 2.
 * Please see the file "COPYING" in this directory for more information
 * about the GNU General Public License Version 2.
 *
 *     Copyright (C) 2009  Kevin Lamonte
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 */

#include "common.h"

/* #define DEBUG_INPUT */
#undef DEBUG_INPUT

#include <unistd.h>
#include "qcurses.h"
#include "qodem.h"
#include "states.h"
#include "status.h"
#include "screensaver.h"
#include "scrollback.h"
#include "phonebook.h"
#include "input.h"

/* How long it's been since user input came in */
static time_t screensaver_time;

#ifdef Q_PDCURSES

static void pdcurses_key(int * key, int * flags) {
	switch (*key) {
	case ALT_0:
		*key = '0';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_1:
		*key = '1';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_2:
		*key = '2';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_3:
		*key = '3';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_4:
		*key = '4';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_5:
		*key = '5';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_6:
		*key = '6';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_7:
		*key = '7';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_8:
		*key = '8';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_9:
		*key = '9';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_A:
		*key = 'a';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_B:
		*key = 'b';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_C:
		*key = 'c';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_D:
		*key = 'd';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_E:
		*key = 'e';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_F:
		*key = 'f';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_G:
		*key = 'g';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_H:
		*key = 'h';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_I:
		*key = 'i';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_J:
		*key = 'j';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_K:
		*key = 'k';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_L:
		*key = 'l';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_M:
		*key = 'm';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_N:
		*key = 'n';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_O:
		*key = 'o';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_P:
		*key = 'p';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_Q:
		*key = 'q';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_R:
		*key = 'r';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_S:
		*key = 's';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_T:
		*key = 't';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_U:
		*key = 'u';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_V:
		*key = 'v';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_W:
		*key = 'w';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_X:
		*key = 'x';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_Y:
		*key = 'y';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_Z:
		*key = 'z';
		*flags = KEY_FLAG_ALT;
		break;
	case CTL_LEFT:
		*key = KEY_LEFT;
		*flags = KEY_FLAG_CTRL;
		break;
	case CTL_RIGHT:
		*key = KEY_RIGHT;
		*flags = KEY_FLAG_CTRL;
		break;
	case CTL_PGUP:
		*key = KEY_PPAGE;
		*flags = KEY_FLAG_CTRL;
		break;
	case CTL_PGDN:
		*key = KEY_NPAGE;
		*flags = KEY_FLAG_CTRL;
		break;
	case CTL_HOME:
		*key = KEY_HOME;
		*flags = KEY_FLAG_CTRL;
		break;
	case CTL_END:
		*key = KEY_END;
		*flags = KEY_FLAG_CTRL;
		break;
	case ALT_MINUS:
		*key = '-';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_EQUAL:
		*key = '=';
		*flags = KEY_FLAG_ALT;
		break;
	case CTL_UP:
		*key = KEY_UP;
		*flags = KEY_FLAG_CTRL;
		break;
	case CTL_DOWN:
		*key = KEY_DOWN;
		*flags = KEY_FLAG_CTRL;
		break;
	case CTL_TAB:
		*key = 0x09;
		*flags = KEY_FLAG_CTRL;
		break;
	case ALT_PGUP:
		*key = KEY_PPAGE;
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_PGDN:
		*key = KEY_NPAGE;
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_HOME:
		*key = KEY_HOME;
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_END:
		*key = KEY_END;
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_UP:
		*key = KEY_UP;
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_DOWN:
		*key = KEY_DOWN;
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_LEFT:
		*key = KEY_LEFT;
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_RIGHT:
		*key = KEY_RIGHT;
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_TAB:
		*key = 0x09;
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_INS:
		*key = KEY_IC;
		*flags = KEY_FLAG_ALT;
		break;
	case CTL_INS:
		*key = KEY_IC;
		*flags = KEY_FLAG_CTRL;
		break;
	case ALT_DEL:
		*key = KEY_DC;
		*flags = KEY_FLAG_ALT;
		break;
	case CTL_DEL:
		*key = KEY_DC;
		*flags = KEY_FLAG_CTRL;
		break;
	case CTL_BKSP:
		*key = KEY_BACKSPACE;
		*flags = KEY_FLAG_CTRL;
		break;
	case ALT_BKSP:
		*key = KEY_BACKSPACE;
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_ENTER:
		*key = KEY_ENTER;
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_ESC:
		*key = KEY_ESCAPE;
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_BQUOTE:
		*key = '`';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_FQUOTE:
		*key = '\'';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_LBRACKET:
		*key = '[';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_RBRACKET:
		*key = ']';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_SEMICOLON:
		*key = ';';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_COMMA:
		*key = ',';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_FSLASH:
		*key = '/';
		*flags = KEY_FLAG_ALT;
		break;
	case ALT_BSLASH:
		*key = '\\';
		*flags = KEY_FLAG_ALT;
		break;
	case CTL_ENTER:
		*key = KEY_ENTER;
		*flags = KEY_FLAG_CTRL;
		break;
	}
} /* ---------------------------------------------------------------------- */

#endif /* Q_PDCURSES */

/*
 * Wrapper for wgetch() that handles the KEY_FLAG_ALT flag appropriately.
 *
 * The keystroke and flags are returned in *keystroke and *flags, respectively.
 *
 * If flags == NULL, KEY_FLAG_ALT is not returned, e.g. Alt-X behaves like X.
 *
 * If usleep_time > 0 then we will sleep that long if wgetch() returns ERR (to
 * prevent a busy wait).
 */
void qodem_win_getch(void * window, int * keystroke, int * flags, const unsigned int usleep_time) {
	time_t current_time;
	int new_height, new_width;
	int param = 0;
	wint_t utf_keystroke;
	Q_BOOL modifier = Q_FALSE;
	int return_keystroke = ERR;

	/* Check for screensaver */
	if (q_screensaver_timeout > 0) {
		time(&current_time);
		if (difftime(current_time, screensaver_time) > q_screensaver_timeout) {
			qlog(_("SCREENSAVER activating...\n"));
			switch_state(Q_STATE_SCREENSAVER);
		}
	}

	/* Assume no KEY_FLAG_ALT or KEY_FLAG_CTRL */
	if (flags != NULL) {
		*flags = 0;
	}

	/* Switch to non-block */
	if (q_keyboard_blocks == Q_TRUE) {
		nodelay(window, FALSE);
	} else {
		nodelay(window, TRUE);
	}

	/* Grab keystroke */
	int res;
	res = wget_wch(window, &utf_keystroke);
	*keystroke = utf_keystroke;
	if ((*keystroke == 0) || (res == ERR)) {
		*keystroke = ERR;
	}

	if (*keystroke != ERR) {
		time(&screensaver_time);
	}

	if (*keystroke == ERR) {
		if ((usleep_time > 0) && (q_keyboard_blocks == Q_FALSE)) {
			usleep(usleep_time);
		}
	} else if ((*keystroke == KEY_RESIZE) && (res == KEY_CODE_YES)) {
		/* Special case: KEY_RESIZE */
		getmaxyx(stdscr, new_height, new_width);

		/*
		 * At this point, the display is hosed.
		 */

		/*
		 * Increase the # of lines to match the new screen size
		 * if necessary.
		 *
		 * Note this check is (new_height - STATUS_HEIGHT), not
		 * (new_height - STATUS_HEIGHT - 1)
		 */
		if (q_status.scrollback_lines < new_height - STATUS_HEIGHT) {
			while (q_status.scrollback_lines < new_height - STATUS_HEIGHT) {
				new_scrollback_line();
			}
			q_scrollback_position = q_scrollback_last;
			q_scrollback_current = q_scrollback_last;

			/* Reset scrolling regions */
			q_status.scroll_region_top = 0;
			q_status.scroll_region_bottom = new_height - STATUS_HEIGHT - 1;

			/*
			 * q_status.cursor_x and cursor_y are almost certainly in
			 * the wrong place.  Let's get them in the bottom left
			 * corner.
			 */
			q_status.cursor_x = 0;
			q_status.cursor_y = new_height - STATUS_HEIGHT - 1;

		} else {
			/*
			 * We have enough scrollback to cover the new size,
			 * we just need to bring q_status.cursor_y down or up.
			 */
			q_status.cursor_y += (new_height - HEIGHT);
			q_status.scroll_region_bottom += (new_height - HEIGHT);
		}
		WIDTH = new_width;
		HEIGHT = new_height;

		if (q_status.cursor_x > WIDTH - 1) {
			q_status.cursor_x = WIDTH - 1;
		}

		/* Fix the phonebook display */
		phonebook_reset();

		q_screen_dirty = Q_TRUE;
		*keystroke = ERR;
	} else if ((*keystroke == KEY_SUSPEND) && (res == KEY_CODE_YES)) {
		/* Special case: KEY_SUSPEND (Ctrl-Z usually) */
		/* For now, map it to Ctrl-Z (decimal 26, hex 1A, ASCII [SUB]) */
		*keystroke = 0x1A;

	} else if ((*keystroke == 0x7F) && (res == OK)) {
		/* Special case: map DEL to KEY_BACKSPACE */
		*keystroke = KEY_BACKSPACE;

#ifdef Q_PDCURSES
	} else if (res == KEY_CODE_YES) {
		/* Handle PDCurses alternate keystrokes */
		pdcurses_key(keystroke, flags);
#endif /* Q_PDCURSES */

	} else if (*keystroke == KEY_ESCAPE) {
		if (flags != NULL) {
			*flags |= KEY_FLAG_ALT;
		}
		/* Grab the next keystroke */
		nodelay(window, TRUE);
		res = wget_wch(window, &utf_keystroke);
		if (q_keyboard_blocks == Q_TRUE) {
			nodelay(window, FALSE);
		} else {
			nodelay(window, TRUE);
		}
		*keystroke = utf_keystroke;
		if ((*keystroke == 0) || (res == ERR)) {
			*keystroke = ERR;
		}
		if (*keystroke == ERR) {
			/* This is actually ESCAPE, not ALT-x */
			if (flags != NULL) {
				*flags &= ~KEY_FLAG_ALT;
			}
			*keystroke = KEY_ESCAPE;
		} else {
			/*
			 * A more complex keyboard sequence has come
			 * in that ncurses doesn't know about.  For
			 * now just use some simple mapping and assume
			 * that all the bytes in the keystroke are
			 * present.
			 */
#ifdef DEBUG_INPUT
			/* DEBUG */
			fprintf(stderr, "1 Keystroke: '%c' %d %o\n",
				*keystroke, *keystroke, *keystroke);
#endif

			if ((*keystroke == '[') && (res == OK)) {
				param = 0;

				while (*keystroke != ERR) {
					/* Grab the next keystroke */
					res = wget_wch(window, &utf_keystroke);
					*keystroke = utf_keystroke;
					if ((*keystroke == 0) || (res == ERR)) {
						*keystroke = ERR;
					}

#ifdef DEBUG_INPUT
					fprintf(stderr, "2 Keystroke: '%c' %d %o\n",
						*keystroke, *keystroke, *keystroke);
#endif
					
					if ((*keystroke >= '0') && (*keystroke <= '9') && (res == OK)) {
						param *= 10;
						param += (*keystroke - '0');
					} else if (((*keystroke == '~') || (*keystroke == ';')) && (res == OK)) {
#ifdef DEBUG_INPUT
						fprintf(stderr, "2 param: %d\n", param);
#endif
						if ((*keystroke == ';') && (res == OK)) {
#ifdef DEBUG_INPUT
							fprintf(stderr, "3 - modifier -\n");
#endif

							/* Param is followed by a modifier */
							modifier = Q_TRUE;
						}

						if ((*keystroke == '~') && (modifier == Q_TRUE) && (res == OK)) {
#ifdef DEBUG_INPUT
							fprintf(stderr, "3 param: %d\n", param);
#endif
							/* Param is a modifier:
							 *
							 * SHIFT 2
							 * ALT   3
							 * CTRL  5
							 */
							switch (param) {
							case 2:
#ifdef DEBUG_INPUT
								fprintf(stderr, "2 SHIFT\n");
#endif
								if (flags != NULL) {
									*flags |= KEY_FLAG_SHIFT;
								}
								break;
							case 3:
#ifdef DEBUG_INPUT
								fprintf(stderr, "2 ALT\n");
#endif
								if (flags != NULL) {
									*flags |= KEY_FLAG_ALT;
								}
								break;
							case 5:
#ifdef DEBUG_INPUT
								fprintf(stderr, "2 CTRL\n");
#endif
								if (flags != NULL) {
									*flags |= KEY_FLAG_CTRL;
								}
								break;
							}
							*keystroke = return_keystroke;
							if ((return_keystroke > 0xFF) && (res == OK)) {
								/* Unicode character */
								*flags |= KEY_FLAG_UNICODE;
							}
							/* Exit the while loop */
							break;
						}


						if (((*keystroke == '~') && (modifier == Q_FALSE) && (res == OK)) ||
							((*keystroke == ';') && (modifier == Q_TRUE) && (res == OK))) {

							switch (param) {
							case 1:
#ifdef DEBUG_INPUT
								fprintf(stderr, "2 Home\n");
#endif
								*keystroke = KEY_HOME;
								break;
							case 4:
#ifdef DEBUG_INPUT
								fprintf(stderr, "2 End\n");
#endif
								*keystroke = KEY_END;
								break;
							case 5:
#ifdef DEBUG_INPUT
								fprintf(stderr, "2 PgUp\n");
#endif
								*keystroke = KEY_PPAGE;
								break;
							case 6:
#ifdef DEBUG_INPUT
								fprintf(stderr, "2 PgDn\n");
#endif
								*keystroke = KEY_NPAGE;
								break;
							default:
								*keystroke = ERR;
								break;
							}

							/* Reset for possible modifier */
							param = 0;

							if ((*keystroke == '~') && (res == OK)) {
								/* Exit the while loop */
								break;
							} else {
								return_keystroke = *keystroke;
								if ((return_keystroke > 0xFF) && (res == OK)) {
									/* Unicode character */
									*flags |= KEY_FLAG_UNICODE;
								}
							}
						}
					}
				}

			} else if ((*keystroke == 'O') && (res == OK)) {
				/* VT100-style function key */
				param = 0;

				/* Grab the next keystroke */
				res = wget_wch(window, &utf_keystroke);
				*keystroke = utf_keystroke;
				if ((*keystroke == 0) || (res == ERR)) {
					*keystroke = ERR;
				}

#ifdef DEBUG_INPUT
				fprintf(stderr, "2 Keystroke: '%c' %d %o\n",
					*keystroke, *keystroke, *keystroke);
#endif

				switch (*keystroke) {
				case 'P':
					*keystroke = KEY_F(1);
					break;
				case 'Q':
					*keystroke = KEY_F(2);
					break;
				case 'R':
					*keystroke = KEY_F(3);
					break;
				case 'S':
					*keystroke = KEY_F(4);
					break;
				case 't':
					*keystroke = KEY_F(5);
					break;
				case 'u':
					*keystroke = KEY_F(6);
					break;
				case 'v':
					*keystroke = KEY_F(7);
					break;
				case 'l':
					*keystroke = KEY_F(8);
					break;
				case 'w':
					*keystroke = KEY_F(9);
					break;
				case 'x':
					*keystroke = KEY_F(10);
					break;
				}
			}
		}
	}

	/* Restore the normal keyboard */
	if (q_keyboard_blocks == Q_TRUE) {
		nodelay(window, FALSE);
	} else {
		nodelay(window, TRUE);
	}

#ifdef DEBUG_INPUT
	/* DEBUG */
	if (*keystroke != -1) {
		if (flags != NULL) {
			fprintf(stderr, "Keystroke: %d %o FLAGS: %02x\n", *keystroke, *keystroke, *flags);
		} else {
			fprintf(stderr, "Keystroke: %d %o FLAGS: NULL\n", *keystroke, *keystroke);
		}
	}
#endif

	/* Set CTRL for normal keystrokes */
	if ((*keystroke < 0x20) && (res == OK)) {
		if (flags != NULL) {
			*flags |= KEY_FLAG_CTRL;
		}
	}

	return;
} /* ---------------------------------------------------------------------- */

void qodem_getch(int * keystroke, int * flags, const unsigned int usleep_time) {
	qodem_win_getch(stdscr, keystroke, flags, usleep_time);
} /* ---------------------------------------------------------------------- */

void discarding_getch() {
	wgetch(stdscr);
} /* ---------------------------------------------------------------------- */

void set_blocking_input(Q_BOOL block) {
	if (block == Q_TRUE) {
		nodelay(stdscr, FALSE);
	} else {
		nodelay(stdscr, TRUE);
	}
} /* ---------------------------------------------------------------------- */
